/* -*- Mode: C; c-basic-offset: 4 -*- */
/* Dia -- a diagram creation/manipulation program
 * Copyright (C) 1998 Alexander Larsson
 *
 * sql-export.c: SQL export filter for Dia
 * Copyright (C) 2001 Andrew Halper
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <glib.h>

#include "message.h"
#include "filter.h"
#include "class.h"
#include "group.h"

/* I don't know what this is.  It is an artifact of dxf-export.c */
#ifdef THIS_IS_PROBABLY_DEAD_CODE
static const gchar *dia_version_string = "Dia-" VERSION;
#endif

static void save_umlclass_table(UMLClass *, FILE *);
static void write_tables(GList *, FILE *);
static void export_sql(DiagramData *, const gchar *, const gchar *,
		       void *);

static void
save_umlclass_table(UMLClass *umlclass, FILE *file)
{
  UMLAttribute *attr;
  GList *list = NULL, *pk = NULL;
  int i, len;
  char *column;

  /* Do not generate DDL for classes that are not explicitly tagged
     "persistent".  This might be too strict to be practical (I can
     see people not liking it), however, it is a nice mechanism to
     allow exclusion of selected tables from the DDL (which I need)
     that is also entirely consistent with UML.  We'll have to wait
     and see how it is recieved.  Note there appears to be some
     disagreement between The Unified Modeling Language User Guide and
     Database Design for Smarties on this.  The Unified Modeling
     Language User Guide says "persistent" is supposed to be a tag,
     and Database Design for Smarties says "persistent" is supposed
     to be a stereotype.  I have gone with The Unified Modeling
     Language User Guide on this one, reasoning that Booch, Rumbaugh,
     and Jacobson wrote the language, they ought to know. */
  if (!strstr(umlclass->name, "{persistent}"))
      return;

  fprintf(file, "CREATE TABLE ");

  /* Table names are in all-uppercase.  It's a Joe Celko SQL style
     thing I like. */
  len = strlen(umlclass->name);
  for (i = 0; i < len && !isspace((int) umlclass->name[i]); i++) {
      fprintf(file, "%c", toupper((int) umlclass->name[i]));
  }

  fprintf(file, " (\n");

  /* Attribute info: */
  list = umlclass->attributes;
  while (list != NULL) {
    attr = (UMLAttribute *) list->data;
    column = attr->name;
    if (strstr(attr->name, "{OID}")) {
	/* squirrel away column name to put in PRIMARY KEY designation below */
	column = strtok(attr->name, " {");
	pk = g_list_append(pk, column);
    }
    fprintf(file, "   %s %s", column, attr->type);
    if ((list = g_list_next(list)) || pk) {
	fprintf(file, ",");
    }
    fprintf(file, "\n");
  }

  if (pk) {
      fprintf(file, "   PRIMARY KEY (");
      list = pk;
      while (list != NULL) {
	  fprintf(file, list->data);
	  if ((list = g_list_next(list))) {
	      fprintf(file, ", ");
	  }
      }
      g_list_free(pk);
      fprintf(file, ")\n");
  }

  fprintf(file, ");\n\n");

} /* end save_umlclass_table */

static void
write_tables(GList *objects, FILE *file)
{
  GList *list;

  list = objects;
  while (list != NULL) {
    Object *obj = (Object *) list->data;

    if IS_GROUP(obj) {
      write_tables(group_objects(obj), file);
    } else {
      /* The SQL export plug-in currently only supports UML diagrams.
	 Someday it might support the ER sheet objects, but this would
	 be considerably more challenging to write (As far as I know,
	 each attribute is a diagram object, and one would need to
	 start following diagram lines around to determine which
	 entities each attribute is part of...does not sound fun, and
	 not worth it IMHO), but since Dia can mix-and-match objects
	 from different sheets on the same diagram, we need to only
	 generate DDL for UML classes only, and ignore all other
	 objects. */
      if (!strcmp(obj->type->name, "UML - Class")) {
	  save_umlclass_table((UMLClass *) obj, file);
      }
    }
    list = g_list_next(list);
  }
} /* end write_tables */

static void
export_sql(DiagramData *data, const gchar *filename, 
           const gchar *diafilename, void* user_data)
{
    FILE *file;
    /* see commented-out relationship code below */
    /* int res; */ 
    int i;
    Layer *layer;

    file = fopen(filename, "w");

    if (file == NULL) {
	  message_error(_("Couldn't open: '%s' for writing.\n"), filename);
	  return;
    }

    fprintf(file, "                /* SQL DDL file written by Dia %s */\n\n",
	    VERSION);
    fprintf(file, "   /* See http://www.lysator.liu.se/~alla/dia/dia.html ");
    fprintf(file, "for more */\n\n");

    /* for each layer */
    for (i = 0; i < data->layers->len; i++) {
	/* write CREATE TABLE statement for objects in layer */
	layer = (Layer *) g_ptr_array_index(data->layers, i);
	write_tables(layer->objects, file);
  
	/* Relationship code generation is not implemented yet.
	   When/if it is, it will probably be better for it to have
	   its own loop, so any relationship DDL appears at the bottom
	   of the file, thereby avoiding forward-referencing errors in
	   some databases, e.g. Ingres */
	/*
	res = write_relationships(layer->objects, file);
	if (!res)
	    return FALSE;
	*/
    }
    fclose(file);
} /* end export_sql */

static const gchar *extensions[] = { "sql", NULL };
DiaExportFilter sql_export_filter = {
    N_("Structured Query Language"),
    extensions,
    export_sql
};
